from contextlib import asynccontextmanager
from datetime import datetime
from typing import (
    AsyncIterator,
    AsyncContextManager,
    Any,
    Callable,
    Iterable,
    Literal,
    Self,
    Sequence,
    Protocol,
    cast,
)
import inspect
import re
import sqlite3
import types
import typing

from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic.fields import PydanticUndefined  # type: ignore
import aiosqlite


def get_pydantic_model(v: Any | None) -> type[BaseModel] | None:
    if inspect.isclass(v) and issubclass(v, BaseModel):
        return v
    if isinstance(v, types.UnionType):
        for variant in typing.get_args(v):
            if (model := get_pydantic_model(variant)) is not None:
                return model
    return None


class TypedCursor[T: BaseModel](Protocol):
    @classmethod
    def wrap_cursor(cls, row_cls: type[T], cursor: aiosqlite.Cursor) -> Self:
        def row_factory(cursor: aiosqlite.Cursor, row: aiosqlite.Row) -> T:
            fields = row_cls.model_fields
            field_names = fields.keys()
            kwargs = dict(zip(field_names, row))
            for key, value in kwargs.items():
                field = fields[key]
                annotation = field.annotation
                if (
                    model := get_pydantic_model(annotation)
                ) is not None and value is not None:
                    kwargs[key] = model.model_validate_json(value)
            return row_cls(**kwargs)

        cursor.row_factory = row_factory  # type: ignore
        return cast(Self, cursor)

    async def __aiter__(self) -> AsyncIterator[T]: ...

    async def execute(self, sql: str, parameters: Iterable[Any] | None) -> Self: ...

    async def executemany(
        self, sql: str, parameters: Iterable[Iterable[Any]]
    ) -> Self: ...

    async def executescript(self, sql_script: str) -> Self: ...

    async def fetchone(self) -> T | None: ...

    async def fetchmany(self, size: int | None = None) -> Iterable[T]: ...

    async def fetchall(self) -> list[T]: ...

    async def close(self) -> None: ...

    @property
    def rowcount(self) -> int: ...

    @property
    def lastrowid(self) -> int | None: ...

    @property
    def arraysize(self) -> int: ...

    @arraysize.setter
    def arraysize(self, value: int) -> None: ...

    @property
    def connection(self) -> sqlite3.Connection: ...

    async def __aenter__(self): ...

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...


@asynccontextmanager
async def select[T: TableModel](
    cls: type[T], db: aiosqlite.Connection, query: str = "", params: Sequence[Any] = ()
) -> AsyncIterator[TypedCursor[T]]:
    field_names = cls.model_fields.keys()
    query = "SELECT {} FROM {} {}".format(
        ", ".join(field_names), cls.__resolved_table_name__, query
    )
    async with db.execute(query, params) as cursor:
        yield TypedCursor.wrap_cursor(cls, cursor)


class TableModel(BaseModel):
    id: int = 0

    __indices__: list[tuple[str, ...]] = []
    __unique__: list[tuple[str, ...]] = []
    __resolved_table_name__: str = ""
    __table_name__: str | None = None

    def __init_subclass__(cls, **kwargs):
        if (table_name := cls.__table_name__) is None:
            table_name = re.sub(r"([a-zA-Z])([A-Z])", r"\1_\2", cls.__name__).lower()
        cls.__resolved_table_name__ = table_name
        super().__init_subclass__(**kwargs)

    @classmethod
    async def sqlite_init(cls, db: aiosqlite.Connection) -> None:
        for subcls in cls.__subclasses__():
            table_name = re.sub(r"([a-zA-Z])([A-Z])", r"\1_\2", subcls.__name__).lower()

            column_names = [name for name in subcls.model_fields.keys()]
            for name in column_names:
                if not re.match(r"^\w+$", name):
                    raise ValueError(f"invalid field name {name!r} on {subcls}")
            column_fields = {
                name: db_field(name, field)
                for name, field in subcls.model_fields.items()
            }

            async with db.execute(f"PRAGMA table_info({table_name})") as cursor:
                table_rows = await cursor.fetchall()

            if table_rows:
                # validate the existing columns
                db_columns: set[str] = set()
                for (
                    cid,
                    name,
                    field_type,
                    notnull,
                    default_value,
                    primary_key,
                    *_,
                ) in table_rows:
                    spec_parts = [
                        field_type.upper(),
                        ("PRIMARY KEY" if primary_key else ""),
                        ("NOT NULL" if notnull else ""),
                        (f"DEFAULT {default_value}" if default_value else ""),
                    ]
                    column_spec = " ".join([x for x in spec_parts if x])
                    db_columns.add(name)
                    if (model_spec := column_fields.get(name)) is not None:
                        if column_spec != model_spec:
                            raise TypeError(
                                f"db column spec does not match model {table_name=} {name=} {column_spec=} != {model_spec=}"
                            )
                    else:
                        raise TypeError(
                            f"db column missing from model {table_name=} {name=}"
                        )

                # create missing columns if possible
                missing_columns = set(column_names) - db_columns
                for name in missing_columns:
                    field = subcls.model_fields[name]
                    if field.default is PydanticUndefined:
                        raise TypeError(
                            f"db column missing and no default value present to create it {table_name=} column={name=}"
                        )
                    alter_query = "ALTER TABLE {} ADD COLUMN {} {}".format(
                        table_name, name, column_fields[name]
                    )
                    _ = await db.execute(alter_query)

            else:
                # create the table
                column_queries = [
                    "{} {}".format(name, field) for name, field in column_fields.items()
                ]

                # unique constraints
                # NOTE: no support for adding new unique constraints?
                for columns in subcls.__unique__:
                    for name in columns:
                        if not re.match(r"^\w+$", name):
                            raise ValueError(
                                f"invalid unique column {name!r} on {subcls}"
                            )
                        if name not in subcls.model_fields:
                            raise ValueError(
                                f"unique on missing column {name!r} on {subcls}"
                            )
                    column_list = ", ".join(columns)
                    column_queries.append(f"UNIQUE({column_list})")
                column_query = ",\n    ".join(column_queries)
                create_query = "CREATE TABLE IF NOT EXISTS {} (\n    {})".format(
                    table_name, column_query
                )
                _ = await db.execute(create_query)

            for columns in subcls.__indices__:
                index_name = "idx_" + ("__".join(columns))
                for name in columns:
                    if not re.match(r"^\w+$", name):
                        raise ValueError(f"invalid index column {name!r} on {subcls}")
                    if name not in subcls.model_fields:
                        raise ValueError(
                            f"index on missing column {name!r} on {subcls}"
                        )
                column_list = ", ".join(columns)
                _ = await db.execute(
                    f"CREATE INDEX IF NOT EXISTS {index_name} ON {table_name}({column_list})"
                )

            await db.commit()

    @classmethod
    async def _create_typed(cls, db: aiosqlite.Connection, obj: Self) -> Self:
        names = [name for name in cls.model_fields.keys() if name != "id"]
        param_str = ", ".join("?" for _ in names)
        name_str = ", ".join(names)
        value_params = [getattr(obj, name) for name in names]
        value_params = [
            v.model_dump_json() if isinstance(v, BaseModel) else v for v in value_params
        ]

        query = f"INSERT INTO {cls.__resolved_table_name__} ({name_str}) VALUES ({param_str})"
        async with db.execute_insert(query, value_params) as rowid_tuple:
            assert rowid_tuple is not None
            obj.id = rowid_tuple[0]
        await db.commit()

        return obj

    @classmethod
    async def create[**P, S: Self](
        cls: Callable[P, S],
        db: aiosqlite.Connection,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> S:
        obj = cls(*args, **kwargs)
        return await obj._create_typed(db, obj)

    @classmethod
    def select(
        cls, db: aiosqlite.Connection, query: str = "", params: Sequence[Any] = ()
    ) -> AsyncContextManager[TypedCursor[Self]]:
        cursor_gen = select(cls, db, query=query, params=params) # type: ignore
        return cast(AsyncContextManager[TypedCursor[Self]], cursor_gen)

    @classmethod
    async def select_count(
        cls, db: aiosqlite.Connection, query: str = "", params: Sequence[Any] = ()
    ) -> int:
        query = "SELECT count(*) FROM {} {}".format(cls.__resolved_table_name__, query)
        async with db.execute(query, params) as cursor:
            row = await cursor.fetchone()
            return row[0] if row is not None else 0

    @classmethod
    async def remove(
        cls, db: aiosqlite.Connection, query: str, params: Sequence[Any] = ()
    ) -> None:
        if "where" not in query.lower():
            raise ValueError("remove() query must contain WHERE clause")
        query = f"DELETE FROM {cls.__resolved_table_name__} {query}"
        await db.execute(query, params)
        await db.commit()

    @classmethod
    async def _update(
        cls,
        db: aiosqlite.Connection,
        query: str,
        params: Sequence[Any] = (),
        /,
        **kwargs,
    ) -> None:
        if "where" not in query.lower():
            raise ValueError("update() query must contain WHERE clause")
        if not kwargs:
            return
        updates = [f"{name}=?" for name in kwargs]
        params = tuple(
            v.model_dump_json() if isinstance(v, BaseModel) else v
            for v in kwargs.values()
        ) + tuple(params)
        query = f"UPDATE {cls.__resolved_table_name__} SET {', '.join(updates)} {query}"
        await db.execute(query, params)
        await db.commit()

    @classmethod
    async def update(
        cls,
        db: aiosqlite.Connection,
        query: str,
        params: Sequence[Any] = (),
        /,
        **kwargs,
    ) -> None:
        cons = cls.model_construct()
        for k, v in kwargs.items():
            cls.__pydantic_validator__.validate_assignment(cons, k, v)
        await cls._update(db, query, params, **kwargs)

    async def update_one(self, db: aiosqlite.Connection, **kwargs: Any) -> None:
        for k, v in kwargs.items():
            self.__pydantic_validator__.validate_assignment(self, k, v)
            setattr(self, k, v)
        await self._update(db, "WHERE id=?", (self.id,), **kwargs)


def db_field(name: str, field: FieldInfo) -> str:
    extra: list[str] = []
    if name == "id":
        extra.append("PRIMARY KEY")

    pytype = field.annotation
    origin = typing.get_origin(pytype)
    optional = False
    if origin is types.UnionType:
        args = typing.get_args(pytype)
        if len(args) == 2 and type(None) in args:
            optional = True
            pytype = [x for x in args if x is not type(None)][0]
    if name != "id" and not optional:
        extra.append("NOT NULL")

    if (default := field.default) is not PydanticUndefined:
        if default is None:
            extra.append("DEFAULT NULL")
        elif isinstance(default, (int, float)):
            extra.append(f"DEFAULT {default}")
        elif isinstance(default, str):
            extra.append(f"DEFAULT {default!r}")
        else:
            raise TypeError(f"unsupported default field {name=} {default=}")

    is_model = inspect.isclass(pytype) and issubclass(pytype, BaseModel)
    if is_model:
        dbtype = "JSONB"
    elif origin is Literal:
        dbtype = "TEXT"
    elif pytype is int:
        dbtype = "INTEGER"
    elif pytype is float:
        dbtype = "REAL"
    elif pytype is bool:
        dbtype = "BOOLEAN"
    elif pytype is str:
        dbtype = "TEXT"
    elif pytype is datetime:
        dbtype = "DATETIME"
    else:
        raise TypeError(f"unknown field type: {pytype}")

    return " ".join([dbtype] + extra)



from datetime import datetime
from typing import Literal
import sqlite3

from pydantic import BaseModel, ValidationError
import aiosqlite
import pytest

from aiosqlite_strict import TableModel, db_field, get_pydantic_model


def test_get_pydantic_model() -> None:
    class Child(BaseModel):
        value: int

    assert get_pydantic_model(Child) is Child
    assert get_pydantic_model(Child | None) is Child
    assert get_pydantic_model(int | None) is None
    assert get_pydantic_model(str) is None


def test_db_field_types_and_defaults() -> None:
    class Nested(BaseModel):
        label: str

    class Fields(BaseModel):
        id: int = 0
        count: int
        ratio: float
        active: bool
        name: str
        created_at: datetime
        kind: Literal["a", "b"]
        note: str | None = None
        nickname: str = "nick"
        meta: Nested

    fields = Fields.model_fields
    assert db_field("id", fields["id"]) == "INTEGER PRIMARY KEY DEFAULT 0"
    assert db_field("count", fields["count"]) == "INTEGER NOT NULL"
    assert db_field("ratio", fields["ratio"]) == "REAL NOT NULL"
    assert db_field("active", fields["active"]) == "BOOLEAN NOT NULL"
    assert db_field("name", fields["name"]) == "TEXT NOT NULL"
    assert db_field("created_at", fields["created_at"]) == "DATETIME NOT NULL"
    assert db_field("kind", fields["kind"]) == "TEXT NOT NULL"
    assert db_field("note", fields["note"]) == "TEXT DEFAULT NULL"
    assert db_field("nickname", fields["nickname"]) == "TEXT NOT NULL DEFAULT 'nick'"
    assert db_field("meta", fields["meta"]) == "JSONB NOT NULL"


def test_db_field_invalid_types() -> None:
    class InvalidDefault(BaseModel):
        bad: int = 1

    class BadDefaults(BaseModel):
        items: int = [1, 2, 3]  # type: ignore[assignment]

    class BadTypes(BaseModel):
        blob: bytes

    with pytest.raises(TypeError):
        db_field("items", BadDefaults.model_fields["items"])

    with pytest.raises(TypeError):
        db_field("blob", BadTypes.model_fields["blob"])

    assert (
        db_field("bad", InvalidDefault.model_fields["bad"])
        == "INTEGER NOT NULL DEFAULT 1"
    )


def test_table_name_resolution() -> None:
    class CamelCase(TableModel):
        name: str

    class CustomName(TableModel):
        __table_name__ = "custom_table"
        name: str

    assert CamelCase.__resolved_table_name__ == "camel_case"
    assert CustomName.__resolved_table_name__ == "custom_table"


@pytest.mark.asyncio
async def test_sqlite_init_and_crud() -> None:
    class Meta(BaseModel):
        tag: str

    class Base(TableModel):
        pass

    class UserProfile(Base):
        __indices__ = [("name",)]
        __unique__ = [("email",)]

        name: str
        email: str
        meta: Meta

    async with aiosqlite.connect(":memory:") as db:
        await Base.sqlite_init(db)

        assert UserProfile.__resolved_table_name__ == "user_profile"

        user1 = await UserProfile.create(
            db, name="name1", email="e1", meta=Meta(tag="t1")
        )
        user2 = await UserProfile.create(
            db, name="name2", email="e2", meta=Meta(tag="t2")
        )

        assert user1.id == 1
        assert user2.id == 2

        async with UserProfile.select(db) as cursor:
            rows = await cursor.fetchall()

        assert len(rows) == 2
        assert isinstance(rows[0], UserProfile)
        assert isinstance(rows[0].meta, Meta)
        assert rows[0].meta.tag == "t1"

        async with UserProfile.select(db, "WHERE name=?", ("name2",)) as cursor:
            row = await cursor.fetchone()

        assert row is not None
        assert row.email == "e2"

        count = await UserProfile.select_count(db)
        assert count == 2

        await user1.update_one(db, name="name1.1")
        assert user1.name == "name1.1"

        await user1.update_one(db)

        with pytest.raises(ValidationError):
            await user1.update_one(db, name=123)  # type: ignore[arg-type]

        async with UserProfile.select(db, "WHERE name=?", ("name1.1",)) as cursor:
            row = await cursor.fetchone()

        assert row is not None
        assert row.email == "e1"

        async with db.execute("PRAGMA index_list(user_profile)") as cursor:
            index_rows = await cursor.fetchall()
        index_names = {row[1] for row in index_rows}
        assert "idx_name" in index_names

        with pytest.raises(sqlite3.IntegrityError):
            await UserProfile.create(db, name="name3", email="e2", meta=Meta(tag="t3"))


@pytest.mark.asyncio
async def test_update_and_remove() -> None:
    class Base(TableModel):
        pass

    class Item(Base):
        name: str
        quantity: int

    async with aiosqlite.connect(":memory:") as db:
        await Base.sqlite_init(db)

        item1 = await Item.create(db, name="first", quantity=1)
        item2 = await Item.create(db, name="second", quantity=2)

        await Item.update(db, "WHERE id=?", [item1.id], name="updated", quantity=3)

        async with Item.select(db, "WHERE id=?", (item1.id,)) as cursor:
            row = await cursor.fetchone()

        assert row is not None
        assert row.name == "updated"
        assert row.quantity == 3

        await Item.remove(db, "WHERE id=?", (item2.id,))
        assert await Item.select_count(db) == 1

        with pytest.raises(ValueError):
            await Item.update(db, "id=?", (item1.id,), name="bad")

        with pytest.raises(ValueError):
            await Item.remove(db, "id=?", (item1.id,))


@pytest.mark.asyncio
async def test_sqlite_init_existing_table_mismatch() -> None:
    class Base(TableModel):
        pass

    class Widget(Base):
        name: int

    async with aiosqlite.connect(":memory:") as db:
        await db.execute(
            "CREATE TABLE widget (id INTEGER PRIMARY KEY, name TEXT NOT NULL)"
        )
        await db.commit()

        with pytest.raises(TypeError):
            await Base.sqlite_init(db)


@pytest.mark.asyncio
async def test_schema_update_add_field_with_default() -> None:
    class Base(TableModel):
        pass

    class Record(Base):
        name: str
        note: str = "n/a"

    async with aiosqlite.connect(":memory:") as db:
        await db.execute(
            "CREATE TABLE record (id INTEGER PRIMARY KEY DEFAULT 0, name TEXT NOT NULL)"
        )
        await db.commit()

        await Base.sqlite_init(db)

        async with db.execute("PRAGMA table_info(record)") as cursor:
            columns = [row[1] for row in await cursor.fetchall()]
        assert "note" in columns


@pytest.mark.asyncio
async def test_schema_update_add_field_without_default_fails() -> None:
    class Base(TableModel):
        pass

    class Record(Base):
        name: str
        note: str

    async with aiosqlite.connect(":memory:") as db:
        await db.execute(
            "CREATE TABLE record (id INTEGER PRIMARY KEY DEFAULT 0, name TEXT NOT NULL)"
        )
        await db.commit()

        with pytest.raises(TypeError):
            await Base.sqlite_init(db)


@pytest.mark.asyncio
async def test_schema_update_rejects_extra_db_column() -> None:
    class Base(TableModel):
        pass

    class Record(Base):
        name: str

    async with aiosqlite.connect(":memory:") as db:
        await db.execute(
            "CREATE TABLE record (id INTEGER PRIMARY KEY DEFAULT 0, name TEXT NOT NULL, extra TEXT NOT NULL)"
        )
        await db.commit()

        with pytest.raises(TypeError):
            await Base.sqlite_init(db)


@pytest.mark.asyncio
async def test_invalid_field_name_is_rejected() -> None:
    class Base(TableModel):
        pass

    _ = type(
        "BadFieldName",
        (Base,),
        {"__annotations__": {"bad-name": str}},
    )

    async with aiosqlite.connect(":memory:") as db:
        with pytest.raises(ValueError):
            await Base.sqlite_init(db)


@pytest.mark.asyncio
async def test_invalid_unique_column_name_is_rejected() -> None:
    class Base(TableModel):
        pass

    class BadUniqueName(Base):
        __unique__ = [("bad-name",)]
        name: str

    async with aiosqlite.connect(":memory:") as db:
        with pytest.raises(ValueError):
            await Base.sqlite_init(db)


@pytest.mark.asyncio
async def test_missing_unique_column_is_rejected() -> None:
    class Base(TableModel):
        pass

    class BadUnique(Base):
        __unique__ = [("missing",)]
        name: str

    async with aiosqlite.connect(":memory:") as db:
        with pytest.raises(ValueError):
            await Base.sqlite_init(db)


@pytest.mark.asyncio
async def test_invalid_index_column_name_is_rejected() -> None:
    class Base(TableModel):
        pass

    class BadIndexName(Base):
        __indices__ = [("bad-name",)]
        name: str

    async with aiosqlite.connect(":memory:") as db:
        with pytest.raises(ValueError):
            await Base.sqlite_init(db)


@pytest.mark.asyncio
async def test_missing_index_column_is_rejected() -> None:
    class Base(TableModel):
        pass

    class BadIndex(Base):
        __indices__ = [("missing",)]
        name: str

    async with aiosqlite.connect(":memory:") as db:
        with pytest.raises(ValueError):
            await Base.sqlite_init(db)
